@@ -1,18 +1,18 @@
 import contextlib
-import hashlib
 from os import path
 import os
 import pathlib
-from typing import Callable, Iterator, TypeVar
+from typing import Callable, Iterator, Optional, TypeVar
 import venv as venv_module
 import subprocess
 import logging
 import tempfile
 
-import requests
 import time
 import dataclasses
 
+from autoupdater import requirements
+
 
 log = logging.getLogger(__name__)
 
@@ -24,7 +24,9 @@ class BaseException(Exception):
 _MIN_TIME_BETWEEN_ATTEMPTS = 10
 
 
-@dataclasses.dataclass(frozen=True, kw_only=True)
+@dataclasses.dataclass(
+    frozen=True,
+)
 class VenvSpec:
     requirements_file: str
     base_directory: pathlib.Path
@@ -39,20 +41,32 @@ class VenvSpec:
         return self.venv_dir() / "bin" / "python"
 
 
-@dataclasses.dataclass(kw_only=True)
+@dataclasses.dataclass()
 class VenvState:
     installed_digest: bytes = b""
     last_updated_timestamp: float = 0
     when_last_update_attempt: float = 0
 
 
-@dataclasses.dataclass(kw_only=True)
+@dataclasses.dataclass()
 class Venv:
     spec: VenvSpec
     state: VenvState
 
+    def approve_venv(self) -> None:
+        with open(self.spec.venv_dir() / "__confirmed_state__", "wb") as f:
+            f.write(self.state.installed_digest)
 
-@dataclasses.dataclass(kw_only=True)
+    def is_venv_approved_for_digest(self, digest: bytes) -> bool:
+        try:
+            with open(self.spec.venv_dir() / "__confirmed_state__", "rb") as f:
+                confirmed_digest = f.read()
+        except FileNotFoundError:
+            return False
+        return confirmed_digest == digest
+
+
+@dataclasses.dataclass()
 class Program:
     process: subprocess.Popen
     venv: Venv
@@ -128,7 +142,7 @@ def run_program_until_dead_or_updated(
     args: list[str],
     duration_between_updates: float,
     termination_timeout: float,
-) -> bytes | None:
+) -> Optional[bytes]:
     with launch(venv, module, args) as program:
         while program.is_running():
             if time.time() - program.when_last_update_check > duration_between_updates:
@@ -183,7 +197,9 @@ def ensure_venv(venv_spec: VenvSpec) -> Venv:
 
 
 def _create_venv(venv_spec: VenvSpec) -> Venv:
-    venv_module.create(venv_spec.venv_dir().absolute(), with_pip=True)
+    venv_module.create(
+        venv_spec.venv_dir().absolute(), with_pip=True, upgrade_deps=True
+    )
     return Venv(spec=venv_spec, state=VenvState())
 
 
@@ -195,8 +211,10 @@ def _recreate_venv(venv_spec: VenvSpec) -> Venv:
     return _create_venv(venv_spec)
 
 
-def maybe_new_requirements_digest(venv: Venv) -> bytes | None:
-    remote_digest = load_requirements_digest(venv.spec.requirements_file)
+def maybe_new_requirements_digest(venv: Venv) -> Optional[bytes]:
+    remote_digest = requirements.digest_from_requirements_file(
+        venv.spec.requirements_file
+    )
     if remote_digest != venv.state.installed_digest:
         return remote_digest
     return None
@@ -204,7 +222,12 @@ def maybe_new_requirements_digest(venv: Venv) -> bytes | None:
 
 def ensure_digest_installed(venv: Venv, target_digest: bytes) -> None:
     if target_digest == venv.state.installed_digest:
-        return venv
+        return
+
+    if venv.is_venv_approved_for_digest(target_digest):
+        venv.state.installed_digest = target_digest
+        venv.state.last_updated_timestamp = time.time()
+        return
 
     time_since_last_attempt = time.time() - venv.state.when_last_update_attempt
     if time_since_last_attempt < _MIN_TIME_BETWEEN_ATTEMPTS:
@@ -213,19 +236,8 @@ def ensure_digest_installed(venv: Venv, target_digest: bytes) -> None:
 
     log.info("Calculating requirements...")
 
-    def no_comments(line: str) -> str:
-        return line.split("#", maxsplit=1)[0].strip()
-
-    def clean_requirement(line: str) -> str:
-        return no_comments(line).split(";", maxsplit=1)[0].strip()
-
     with open(venv.spec.requirements_file, "r") as f:
-        target_requirements = [
-            (clean_requirement(line), no_comments(line))
-            for line in f.readlines()
-            if line.strip()
-        ]
-        clean_target_requirements = [r[0] for r in target_requirements]
+        requirements_content = f.read()
 
     pip_freeze = subprocess.run(
         [venv.spec.pip_path().absolute(), "freeze"],
@@ -234,19 +246,9 @@ def ensure_digest_installed(venv: Venv, target_digest: bytes) -> None:
         text=True,
     )
 
-    installed_requirements = [
-        line.strip() for line in pip_freeze.stdout.split("\n") if line.strip()
-    ]
-
-    requirements_to_remove = [
-        line for line in installed_requirements if line not in clean_target_requirements
-    ]
-
-    requirements_to_install = [
-        line
-        for clean_requirement, line in target_requirements
-        if clean_requirement not in installed_requirements
-    ]
+    requirements_to_remove, requirements_to_install = requirements.diff(
+        pip_freeze.stdout, requirements_content
+    )
 
     log.info(
         "Remove:\n%s\n\nInstall:\n%s",
@@ -291,35 +293,4 @@ def ensure_digest_installed(venv: Venv, target_digest: bytes) -> None:
     # In that case the update will be triggered again, but will be mostly noop
     venv.state.installed_digest = target_digest
     venv.state.last_updated_timestamp = time.time()
-
-
-def load_requirements_digest(requirements_file: str) -> bytes | None:
-    if requirements_file.startswith("https://") or requirements_file.startswith(
-        "http://"
-    ):
-        data = _load_file_from_web(requirements_file)
-    else:
-        try:
-            with open(requirements_file, "rb") as file_:
-                data = file_.read()
-        except OSError:
-            data = None
-    if data is None:
-        return None
-    return hashlib.sha256(data).digest()
-
-
-def _load_file_from_web(requirements_file: str, retries: int = 10) -> bytes | None:
-    for try_ in range(retries):
-        response = requests.get(requirements_file)
-        if response.status_code == 200:
-            break
-        time.sleep(30)
-    else:
-        log.error(
-            "Could not load the requirements from %s: Status code %s",
-            requirements_file,
-            response.status_code,
-        )
-        return None
-    return response.content
+    venv.approve_venv()
